#include "hbookoo.h"
#include "hbook.h"
#include<iostream.h>
#include<vector>
#include<map>
#include<algorithm>

//**********************************************************************
// class HBEntity
//**********************************************************************

HBEntity *HBEntity::makeObject (int theID) {
  // this method creates an object of a class derived from HBEntity
  // for an id of something that is already in memory.
  HBEntity *result = 0;
  
  //Check different possible types
  FInteger kind = ::hkind (theID);
  switch (kind) {
    case 0: // histogram does not exist
      cerr << "HBEntity::makeObject: Cannot create object for id = " << theID
	    << ", does not exist!\n";
      break;
    case  1: // one-dimensional plot
      cout << "HBEntity::makeObject: Creating a HBHist1 object for  id = " << theID 
	   << ".\n";
      result = HBHist1::makeObject (theID);
      break;
    case  2: // two-dimensional plot
    case  3: // table
    case  4: // Ntuple
    case  8: // profile histogram
    case -1: // unknown kind of histogram
    default:
      cerr << "HBEntity::makeObject: Cannot create object for id = " << theID
	    << ", don't know what kind = " << kind << " means.\n";
      break;
  }  
    
  return result; 
}

HBEntity::MapType HBEntity::entityList; 

HBEntity::HBEntity (int newID) {
  id = newID;
  entityList [id] = this;
//  cerr << "HBEntity::HBEntity: New HBEntity has been created, id=" << id << "\n";
}

HBEntity *HBEntity::getHBEntity (int theID) {
  return entityList[theID];
}

HBEntity::~HBEntity() {
  entityList.erase (id);
  cerr << "HBEntity::~HBEntity: HBEntity is being deleted, id=" << id << "\n";
}

//**********************************************************************
// class HBHisto
//**********************************************************************

HBHisto::HBHisto (int theID) 
  : HBEntity (theID) 
{
//  cerr << "HBHisto::HBHisto: New HBHisto has been created, id=" << getID() << "\n";
}

HBHisto::~HBHisto () 
{
  hdelet (getID());
  cerr << "HBHisto::~HBHisto: HBHisto has been destroyed, id=" << getID() << "\n";
}

HBHisto *HBHisto::hrin (int theID, int icycle, int ioffset) {
  if (theID <= 0) {
    cerr << "HBHisto::hrin: id must be a positive number!\n";
    return 0;	     
  }
  if (getHBEntity (theID+ioffset) != 0) {
    cerr << "HBHisto::hrin: Histogram cannot be read in, id=" << theID 
	 << ", offset=" << ioffset << ", already in memoty!\n";
    return 0;	      // ID already in use
  }
  ::hrin (theID, icycle, ioffset);
  return dynamic_cast<HBHisto *> (HBEntity::makeObject (theID+ioffset));
}
  
  
void HBHisto::hbarx() 
{
  ::hbarx (getID());
}

//**********************************************************************
// class HBFunctionAdaptorBase
//**********************************************************************

void *HBFunctionAdaptorBase::fobject = 0;

//**********************************************************************
// class HBHist1
//**********************************************************************

HBHist1 *HBHist1::hcopy (int newID, const FCharacter *chtitle) {
  if (getHBEntity (newID) != 0) {
    cerr << "HBHist1::hcopy: HBHist1 cannot be created, id=" << newID << "\n";
    return 0;         // ID already in use
  } else {
    cerr << "HBHist1::hcopy: HBHist1 will be created, id=" << newID << "\n";
    return new HBHist1 (*this, newID, chtitle);
  }
}  
  
HBHist1 *HBHist1::hbook1 (int theID, 
 			  const FCharacter *chtitle,
 			  int nx,
 			  float xmi,
 			  float xma,
 			  float vmx) {
  if (getHBEntity (theID) != 0) {
    cerr << "HBHist1::hbook1: HBHist1 cannot be created, id=" << theID << "\n";
    return 0;         // ID already in use
  } else {
//    cerr << "HBHist1::hbook1: HBHist1 will be created, id=" << theID << "\n";
    return new HBHist1 (theID, chtitle, nx, xmi, xma, vmx);
  }
}

HBHist1 *HBHist1::hbfun1 (int theID, 
 			  const FCharacter *chtitle,
 			  int nx,
 			  float xmi,
 			  float xma,
 			  FReal fun (const FReal&)) {
  if (getHBEntity (theID) != 0) {
    cerr << "HBHist1::hbfun1: HBHist1 cannot be created, id=" << theID << "\n";
    return 0;         // ID already in use
  } else {
    cerr << "HBHist1::hbfun1: HBHist1 will be created, id=" << theID << "\n";
    return new HBHist1 (theID, chtitle, nx, xmi, xma, fun);
  }
}

HBHist1 *HBHist1::hbookb (int theID, 
           		  const FCharacter *chtitle,
           		  int nx,
           		  const FReal xbins[],
           		  float vmx) {
  if (getHBEntity (theID) != 0) {
    cerr << "HBHist1::hbookb: HBHist1 cannot be created, id=" << theID << "\n";
    return 0;         // ID already in use
  } else {
//    cerr << "HBHist1::hbookb: HBHist1 will be created, id=" << theID << "\n";
    return new HBHist1 (theID, chtitle, nx, xbins, vmx);
  }
}

HBHist1 *HBHist1::makeObject (int theID) {
  // this method creates an object of class HBHist1
  // for an id of something that is already in memory.
  HBHist1 *result = 0;
  
  //Check different possible types
  FInteger kind = ::hkind (theID);
  switch (kind) {
    case 0: // histogram does not exist
      cerr << "HBHist1::makeObject: Cannot create object for id = " << theID
	    << ", does not exist!\n";
      break;
    case  1: // one-dimensional plot
      cout << "HBHist1::makeObject: Creating a HBHist1 object for  id = " << theID 
	   << ".\n";
      {
      	HBEntity *oldObject = getHBEntity (theID);
      	if (oldObject != 0) {
	  // if already an objet corresponding to this ID exists,
	  // check whether it is of the correct type.
	  if ((result = dynamic_cast<HBHist1 *> (oldObject)) == 0) {
	    // oldObject id NOT a one-dimensional histogram => delete it,
	    // it cannot be used anymore anyway
	    cerr << "HBHist1::makeObject: Deleting old object for id = " << theID
		 << ", it was not a one-dimensional histogram.\n";
	    delete oldObject;
	    // now create a new HBHist1 object
	    result = new HBHist1 (theID);
	  }
	  else {
	    cerr << "HBHist1::makeObject: Old object for id = " << theID
		 << " was a one-dimensional histogram, keep it.\n";
	  }
	}
	else {
	  cerr << "HBHist1::makeObject: No old object for id = " << theID
	       << " present, creating new one.\n";
	  result = new HBHist1 (theID);
	}
      }
      break;
    default:
      cerr << "HBHist1::makeObject: Cannot create object for id = " << theID
	    << ", wrong kind = " << kind << ".\n";
      break;
  }  
    
  return result; 
}
  
HBHist1::HBHist1 (int theID, 
 		  const FCharacter *chtitle,
 		  int nx,
 		  float xmi,
 		  float xma,
 		  float vmx) 
  : HBHisto (theID) {
  ::hbook1 (getID(), chtitle, nx, xmi, xma, vmx);
//  cerr << "HBHist1::HBHist1: New HBHist1 has been created, id=" << getID() << "\n";
//  cerr << "Title: " << chtitle << "\n";
//  cerr << "xmi=" << xmi << ", xma=" << xma << ", vmx=" << vmx << "\n";
}

HBHist1::HBHist1 (int theID, 
 		  const FCharacter *chtitle,
 		  int nx,
 		  float xmi,
 		  float xma,
 		  FReal fun (const FReal&)) 
  : HBHisto (theID) {
  ::hbfun1 (getID(), chtitle, nx, xmi, xma, fun);
//   cerr << "HBHist1::HBHist1: New HBHist1 has been created, id=" << getID() << "\n";
//   cerr << "Title: " << chtitle << "\n";
//   cerr << "Will be filled with a function.\n";
}

HBHist1::HBHist1 (int theID, 
           	  const FCharacter* chtitle,
           	  int nx,
           	  const FReal xbins[],
           	  float vmx) 
  : HBHisto (theID) 
{
  ::hbookb (getID(), chtitle, nx, xbins, vmx);
//   cerr << "HBHist1::HBHist1: New HBHist1 has been created, id=" << getID() << "\n";
//   cerr << "Title: " << chtitle << "\n";
//   cerr << "with the following bin edges:\n";
//  for (int i = 0; i <= nx; ++i) {
//    cerr << xbins [i] << endl;
//  }
} 

HBHist1::HBHist1 (const HBHist1 &oldHist1,
                  int theID, 
                  const FCharacter *chtitle) 
  : HBHisto (theID) {
  ::hcopy (oldHist1.getID(), getID(), chtitle);
//   cerr << "HBHist1::HBHist1: New HBHist1 has been created, id=" << getID() << "\n";
//   cerr << "Title: " << chtitle << "\n";
//   cerr << "from histogram " << oldHist1.getID() << "\n";
}

HBHist1 *HBHist1::getHBHist1 (int theID) {
  return dynamic_cast<HBHist1 *> (getHBEntity (theID));
}

void HBHist1::hfill (float x, float weight) {
  ::hfill (getID(), x, 0, weight); 
}

void HBHist1::hbar() 
{
  ::hbarx (getID());
}

//**********************************************************************
// class HBHist2
//**********************************************************************

HBHist2 *HBHist2::hcopy (int newID, const FCharacter *chtitle) {
  if (getHBEntity (newID) != 0) {
    cerr << "HBHist2::hcopy: HBHist2 cannot be created, id=" << newID << "\n";
    return 0;         // ID already in use
  } else {
    cerr << "HBHist2::hcopy: HBHist2 will be created, id=" << newID << "\n";
    return new HBHist2 (*this, newID, chtitle);
  }
}  
  
HBHist2 *HBHist2::hbook2 (int theID, 
 				 const FCharacter *chtitle,
 				 int nx,
 				 float xmi,
 				 float xma,
 				 int ny,
 				 float ymi,
 				 float yma,
 				 float vmx) {
  if (getHBEntity (theID) != 0) {
    cerr << "HBHist2::hbook2: HBHist2 cannot be created, id=" << theID << "\n";
    return 0;         // ID already in use
  } else {
//    cerr << "HBHist2::hbook2: HBHist2 will be created, id=" << theID << "\n";
    return new HBHist2 (theID, chtitle, nx, xmi, xma, ny, ymi, yma, vmx);
  }
}

HBHist2 *HBHist2::hbfun2 (int theID, 
 				 const FCharacter *chtitle,
 				 int nx,
 				 float xmi,
 				 float xma,
 				 int ny,
 				 float ymi,
 				 float yma,
 				 FReal fun (const FReal&, const FReal&)) {
  if (getHBEntity (theID) != 0) {
    cerr << "HBHist2::hbfun2: HBHist2 cannot be created, id=" << theID << "\n";
    return 0;         // ID already in use
  } else {
    cerr << "HBHist2::hbfun2: HBHist2 will be created, id=" << theID << "\n";
    return new HBHist2 (theID, chtitle, nx, xmi, xma, ny, ymi, yma, fun);
  }
}

HBHist2::HBHist2 (int theID, 
 		  const FCharacter *chtitle,
 		  int nx,
 		  float xmi,
 		  float xma,
 		  int ny,
 		  float ymi,
 		  float yma,
 		  float vmx) 
  : HBHisto (theID) 
{
  ::hbook2 (getID(), chtitle, nx, xmi, xma, ny, ymi, yma, vmx);
//  cerr << "HBHist2::HBHist2: New HBHist2 has been created, id=" << getID() << "\n";
//  cerr << "Title: " << chtitle << "\n";
//  cerr << "xmi=" << xmi << ", xma=" << xma << "\n";
//  cerr << "ymi=" << ymi << ", yma=" << yma << ", vmx=" << vmx << "\n";
}

HBHist2::HBHist2 (int theID, 
 		  const FCharacter *chtitle,
 		  int nx,
 		  float xmi,
 		  float xma,
 		  int ny,
 		  float ymi,
 		  float yma,
 		  FReal fun (const FReal&, const FReal&)) 
  : HBHisto (theID) {
  ::hbfun2 (getID(), chtitle, nx, xmi, xma, ny, ymi, yma, fun);
  cerr << "HBHist2::HBHist2: New HBHist2 has been created, id=" << getID() << "\n";
  cerr << "Title: " << chtitle << "\n";
  cerr << "Will be filled with function values.\n";
}

HBHist2::HBHist2 (const HBHist2 &oldHist2,
                  int theID, 
                  const FCharacter *chtitle) 
  : HBHisto (theID) {
  ::hcopy (oldHist2.getID(), getID(), chtitle);
  cerr << "HBHist2::HBHist2: New HBHist2 has been created, id=" << getID() << "\n";
  cerr << "Title: " << chtitle << "\n";
  cerr << "from histogram " << oldHist2.getID() << "\n";
}

HBHist2 *HBHist2::getHBHist2 (int theID) {
  return dynamic_cast<HBHist2 *> (getHBEntity (theID));
}

void HBHist2::hfill (float x, float y, float weight) {
  ::hfill (getID(), x, y, weight); 
}

HBHist2 *HBHist2::makeObject (int theID) {
  // this method creates an object of class HBHist1
  // for an id of something that is already in memory.
  HBHist2 *result = 0;
  
  //Check different possible types
  FInteger kind = ::hkind (theID);
  switch (kind) {
    case 0: // histogram does not exist
      cerr << "HBHist2::makeObject: Cannot create object for id = " << theID
	    << ", does not exist!\n";
      break;
    case  2: // two-dimensional plot
      cout << "HBHist2::makeObject: Creating a HBHist2 object for  id = " << theID 
	   << ".\n";
      {
      	HBEntity *oldObject = getHBEntity (theID);
      	if (oldObject != 0) {
	  // if already an objet corresponding to this ID exists,
	  // check whether it is of the correct type.
	  if ((result = dynamic_cast<HBHist2 *> (oldObject)) == 0) {
	    // oldObject id NOT a two-dimensional histogram => delete it,
	    // it cannot be used anymore anyway
	    cerr << "HBHist2::makeObject: Deleting old object for id = " << theID
		 << ", it was not a two-dimensional histogram.\n";
	    delete oldObject;
	    // now create a new HBHist2 object
	    result = new HBHist2 (theID);
	  }
	  else {
	    cerr << "HBHist2::makeObject: Old object for id = " << theID
		 << " was a two-dimensional histogram, keep it.\n";
	  }
	}
	else {
	  cerr << "HBHist2::makeObject: No old object for id = " << theID
	       << " present, creating new one.\n";
	  result = new HBHist2 (theID);
	}
      }
      break;
    default:
      cerr << "HBHist2::makeObject: Cannot create object for id = " << theID
	    << ", wrong kind = " << kind << ".\n";
      break;
  }  
    
  return result; 
}
  
void HBHist2::hbar() 
{
  ::hbar2 (getID());
}

void HBHist2::hbary() 
{
  ::hbary (getID());
}

void HBHist2::hbar2() 
{
  ::hbar2 (getID());
}


//**********************************************************************
// class HBFile
//**********************************************************************

HBFile::MapType HBFile::fileMap;
