#ifndef HBOOKOO_H
#define HBOOKOO_H
#include <map>
#include <vector>

#ifndef HBOOK_H
#  include "hbook.h"
#endif

#ifndef FORTCLOSE_H
#  include "fortclose.h"
#endif 

#ifndef NHMEM_H
#  include "nhmem.h"
#endif

#include <iterator>
#include <limits.h>
#include <iostream>
#include <string>

class HBEntity;
//**********************************************************************
class HBEntity {
//**********************************************************************
public:
  int getID () const;
  static HBEntity *getHBEntity (int theID);
  static inline void htitle (const FCharacter *chgtit);
  
  static HBEntity *makeObject (int theID);
  
  typedef map<int, HBEntity *> MapType;
  typedef MapType::iterator MapIterator;
  
  virtual ~HBEntity();

protected:
  explicit HBEntity (int newID);  
   
private:
  HBEntity (const HBEntity&);                    // no copy constructor
  const HBEntity& operator= (const HBEntity&);   // no assignment
  
  static MapType entityList;

// Data members:
  int id;
};


//**********************************************************************
class HBHisto: public HBEntity {
//**********************************************************************
public:
  virtual ~HBHisto();

  virtual HBHisto *hcopy (int newID, const FCharacter *chtitle) = 0;

  inline void hprint() const;
  
  virtual void hbar() = 0;
  virtual void hbarx();
  
  static HBHisto *hrin (int theID, int icycle = 999, int ioffset = 0);
  
  static inline void histdo();
  
protected:
  explicit HBHisto (int newID);

private:
  const HBHisto& operator= (const HBHisto&);     // no assignment
};

class HBHist1;                          
//**********************************************************************
class HBFunctionAdaptorBase {
//**********************************************************************

// This is a helper class that stores any pointer of type
// F * in a static variable, and returns it upon request.
// The return type of getFunction is chosen by the dummy
// argument, which is conveniently given by
//   F *f = HBFunctionAdaptorBase::getFunction ((F *)(0));
// i.e. 0 is made into a pointer to F.
// This clase is used by HBFunctionAdaptor1<F> and HBFunctionAdaptor2<F>;
// these classes cannot have static variables of their own,
// because for each specialization of these templates the static variable
// would have to be defined in some compilation unit.
//
// This class also encapsulates the knowledge how the pointer
// is actually stored, so that no casts are necessary
// outside this class.
//
// HBFunctionAdaptor1<F> and HBFunctionAdaptor2<F> must (privately)
// inherit from HBFunctionAdaptorBase, to stress that this class is not
// meant as "general purpose" class.

protected:
  
  template<class F>
  static void setFunction (const F *f) {
    fobject = reinterpret_cast<void *> (f);
  }
  template<class F>
  static F *getFunction(const F *) {
    return reinterpret_cast<F *> (fobject);
  }
  
private:
  static void *fobject;
  HBFunctionAdaptorBase() {}
};

//**********************************************************************
template<class F> 
class HBFunctionAdaptor1: private HBFunctionAdaptorBase {
//**********************************************************************
public:
  static void setFunction (const F *f) {HBFunctionAdaptorBase::setFunction (f);}
  static void setFunction (const F& f) {HBFunctionAdaptorBase::setFunction (&f);}
  static FReal fun (const FReal& x) {
    return static_cast<FReal> ((*HBFunctionAdaptorBase::getFunction((F *)(0))) (x));
  }
};

//**********************************************************************
template<class SomeHBHist1Iterator>
class HBHist1IteratorBase {
//**********************************************************************

// General iterator base class for one-dimensional histograms
//
// The specific iterator classes inherit from this class,
// this is done to avoid the necessity of defining
// all possible increment and decrement operators 
// for each iterator class


public:
  ~HBHist1IteratorBase() {}
  
  inline SomeHBHist1Iterator& operator= (const SomeHBHist1Iterator& rhs);
  
// Some typedefs needed by the STL

  typedef SomeHBHist1Iterator             iterator;
  typedef SomeHBHist1Iterator             const_iterator;
  typedef SomeHBHist1Iterator             self;

  typedef random_access_iterator_tag      iterator_category;
  typedef FReal                           value_type;
  typedef SomeHBHist1Iterator *           pointer;
  typedef SomeHBHist1Iterator&            reference;
//  typedef                               link_type;  // ???
  typedef int                             size_type;
  typedef int                             difference_type;
  
  inline SomeHBHist1Iterator& operator+= (int offset);
  inline SomeHBHist1Iterator& operator-= (int offset);
  inline SomeHBHist1Iterator  operator+  (int offset);
  inline SomeHBHist1Iterator  operator-  (int offset);
  inline SomeHBHist1Iterator& operator++ ();	        //prefix increment
  inline SomeHBHist1Iterator& operator-- ();	        //prefix decrement
  inline SomeHBHist1Iterator  operator++ (int);         //postfix increment
  inline SomeHBHist1Iterator  operator-- (int);         //postfix decrement
  
  inline bool operator== (int index); 
  inline bool operator!= (int index); 
  inline bool operator<  (int index); 
  inline bool operator<= (int index); 
  inline bool operator>  (int index); 
  inline bool operator>= (int index); 
    
//   friend int  operator-  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   
//   friend bool operator== (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   friend bool operator!= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   friend bool operator<  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   friend bool operator<= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   friend bool operator>  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
//   friend bool operator>= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs);
  
protected:
  explicit HBHist1IteratorBase (HBHist1& hist, int column = 1) 
  : theHistogram (hist), theColumn (column)
  {}
  
  HBHist1IteratorBase (const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs) 
  : theHistogram (rhs.theHistogram), theColumn (rhs.theColumn)
  {}
  
  HBHist1& theHistogram;
  int theColumn;
};

//**********************************************************************
class HBHist1Iterator: public HBHist1IteratorBase<HBHist1Iterator> {
//**********************************************************************

// General iterator class for one-dimensional histograms


public:
  explicit HBHist1Iterator (HBHist1& hist, int column = 1) 
  : HBHist1IteratorBase<HBHist1Iterator> (hist, column) 
  {}
  
  HBHist1Iterator (const HBHist1IteratorBase<HBHist1Iterator>& rhs) 
  : HBHist1IteratorBase<HBHist1Iterator> (rhs)
  {}
  
  ~HBHist1Iterator() {}
};

//**********************************************************************
class HBHist1ContentsIterator: 
public HBHist1IteratorBase<HBHist1ContentsIterator> {
//**********************************************************************

// This class provides iterators that can be used to iterate
// through the contents of a one-dimensional histogram.

public:

  explicit HBHist1ContentsIterator (HBHist1& hist, int column = 1) 
  : HBHist1IteratorBase<HBHist1ContentsIterator> (hist, column)
  {}
  
  HBHist1ContentsIterator (const HBHist1IteratorBase<HBHist1ContentsIterator>& rhs) 
  : HBHist1IteratorBase<HBHist1ContentsIterator> (rhs)
  {}
  
  ~HBHist1ContentsIterator() {}
  
  inline value_type operator[](difference_type offset);
  
  inline value_type operator*();
};

//**********************************************************************
class HBHist1Contents {
//**********************************************************************

// This class behaves as if an object of type HBHist1Contents
// was a container for the contents of a one-dimensional histogram

public:
  HBHist1Contents (HBHist1& hist)
  : theHistogram (hist)
  {}
  
// some typedefs, analogous to those defined by the vector template
  typedef HBHist1ContentsIterator iterator;
//  typedef const HBHist1ContentsIterator const_iterator;
  typedef HBHist1ContentsIterator const_iterator;

  typedef iterator::value_type value_type;
  typedef value_type* pointer;
  typedef const value_type* const_pointer;
  typedef value_type& reference;
  typedef const value_type& const_reference;
  typedef iterator::size_type size_type;
  typedef iterator::difference_type difference_type;

  
  inline HBHist1ContentsIterator begin();
  inline HBHist1ContentsIterator end();
  
  inline difference_type size();
  
  inline value_type operator[](difference_type offset);
  
  template<class Container>
  inline HBHist1Contents& operator= (Container rhs);

private:
  HBHist1& theHistogram;
};

//**********************************************************************
class HBHist1ErrorsIterator: 
public HBHist1IteratorBase<HBHist1ErrorsIterator> {
//**********************************************************************

// This class provides iterators that can be used to iterate
// through the Errors of a one-dimensional histogram.

public:

  explicit HBHist1ErrorsIterator (HBHist1& hist, int column = 1) 
  : HBHist1IteratorBase<HBHist1ErrorsIterator> (hist, column)
  {}
  
  HBHist1ErrorsIterator (const HBHist1IteratorBase<HBHist1ErrorsIterator>& rhs) 
  : HBHist1IteratorBase<HBHist1ErrorsIterator> (rhs)
  {}
  
  ~HBHist1ErrorsIterator() {}
  
  inline value_type operator[](difference_type offset);
  
  inline value_type operator*();
};

//**********************************************************************
class HBHist1Errors {
//**********************************************************************

// This class behaves as if an object of type HBHist1Errors
// was a container for the Errors of a one-dimensional histogram

public:
  HBHist1Errors (HBHist1& hist)
  : theHistogram (hist)
  {}
  
// some typedefs, analogous to those defined by the vector template
  typedef HBHist1ErrorsIterator iterator;
//  typedef const HBHist1ErrorsIterator const_iterator;
  typedef HBHist1ErrorsIterator const_iterator;

  typedef iterator::value_type value_type;
  typedef value_type* pointer;
  typedef const value_type* const_pointer;
  typedef value_type& reference;
  typedef const value_type& const_reference;
  typedef iterator::size_type size_type;
  typedef iterator::difference_type difference_type;

  
  inline HBHist1ErrorsIterator begin();
  inline HBHist1ErrorsIterator end();
  
  inline difference_type size();
  
  inline value_type operator[](difference_type offset);
  
  template<class Container>
  inline HBHist1Errors& operator= (Container rhs);

private:
  HBHist1& theHistogram;
};

//**********************************************************************
class HBHist1: public HBHisto {
//**********************************************************************
public:
  virtual HBHist1 *hcopy (int newID, 
                          const FCharacter *chtitle);
                          
                          
  inline float hrndm1() const;                        

  static HBHist1 *hbook1 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          float vmx = 0);
                          
  static HBHist1 *hbfun1 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          FReal fun (const FReal&));
                          
  static HBHist1 *makeObject (int theID);
  
                         
  template<class F>                        
  static HBHist1 *hbfun1 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          F fobject) {
    if (getHBEntity (theID) != 0) {
      cerr << "HBHist1::hbfun1: HBHist1 cannot be created, id=" << theID << "\n";
      return 0; 	// ID already in use
    } else {
      cerr << "HBHist1::hbfun1: HBHist1 will be created, id=" << theID << "\n";
      HBFunctionAdaptor1<F>::setFunction (fobject);     // this function sets static
                                                        // pointer F in class Function<F>
                                                        // to fobject, which is the then
                                                        // called by Function<F>::fun
      return new HBHist1 (theID, chtitle, nx, xmi, xma, &HBFunctionAdaptor1<F>::fun);
    }
  }
  
  static HBHist1 *hbookb (int theID, 
           		  const FCharacter *chtitle,
           		  int nx,
           		  const FReal xbins[],
           		  float vmx = 0);

  template<class Iterator>
  static HBHist1 *hbookb (int theID, 
           		  const FCharacter *chtitle,
           		  Iterator begin,
           		  Iterator end,
           		  float vmx = 0) {
    HBHist1 *hist = 0;
    typename iterator_traits<Iterator>::difference_type n = 0; 
    distance (begin, end, n);  // Don't forget to set n to 0!!!
    n -= 1;
    if (n < 2) {
      cerr << "HBHist1::hbookb: Cannot create histogram, too few bins: n = "
           << n << endl;
    }
    else if (n > 1000) {
      cerr << "HBHist1::hbookb: Cannot create histogram, too many bins: n = "
           << n << endl;
    }
    else {
//      cerr << "HBHist1::hbookb: Will create histogram, n = " << n << endl;
      FReal *bins = new FReal [n];
      FReal *out = bins;
      for (Iterator in = begin; in != end; ++in, ++out) {
    	*out = *in;
      }
      hist = hbookb (theID, chtitle, n, bins, vmx);
      delete[] bins;
    }
    return hist;
  }
  
  template<class T>
  static HBHist1 *hbookb (int theID, 
           		  const FCharacter *chtitle,
           		  int nx,
           		  const vector<T>& xbins,
           		  float vmx = 0) {
    int n = min (nx, int (xbins.size()));       		  
    return hbookb (theID, chtitle, xbins.begin(), xbins.begin()+n+1, vmx);
  }

  template<class T>
  static HBHist1 *hbookb (int theID, 
           		  const FCharacter *chtitle,
           		  const vector<T>& xbins,
           		  float vmx = 0) {
    return hbookb (theID, chtitle, xbins.begin(), xbins.end(), vmx);
  }
           		  
  HBHist1 *hfpak1 (FInteger& nid,
  		   const FReal v[],
  		   int n) {
    ::hfpak1 (getID(), nid, v, n);
    return this;
  }

  HBHist1 *hfpak1 (const FReal v[],
  		   int n) {
    ::hfpak1 (getID(), v, n);
    return this;
  }

  template<class Iterator>
  HBHist1 *hfpak1 (FInteger& nid,
  		   Iterator begin,
  		   Iterator end) {
    typename iterator_traits<Iterator>::difference_type n = 0; 
    distance (begin, end, n);  // Don't forget to set n to 0!!!
    n -= 1;
    if (n >= 1) {
      FReal *bins = new FReal [n];
      FReal *out = bins;
      for (Iterator in = begin; in != end; ++in, ++out) {
    	*out = *in;
      }
      hfpak1 (nid, bins, n);
      delete[] bins;
    }
    return this;
  }
  
  template<class Iterator>
  HBHist1 *hfpak1 (Iterator begin,
  		   Iterator end) {
    FInteger nid = 0;		   
    return hfpak1 (nid, begin, end);
  }

  HBHist1 *hfpak1 (FInteger& nid,
  		   const vector<FReal>& v,
  		   int n) {
    return hfpak1 (nid, v.begin(), v.begin()+n+1);
  }

  HBHist1 *hfpak1 (const vector<FReal>& v,
  		   int n) {
    return hfpak1 (v.begin(), v.begin()+n+1);
  }

  HBHist1 *hfpak1 (FInteger& nid,
  		   const vector<FReal>& v) {
    return hfpak1 (nid, v.begin(), v.end());
  }

  HBHist1 *hfpak1 (const vector<FReal>& v) {
    return hfpak1 (v.begin(), v.end());
  }
    
    
  HBHist1 *hipak1 (FInteger& nid,
  		   const FInteger iv[],
  		   int n) {
    ::hipak1 (getID(), nid, iv, n);
    return this;
  }

  HBHist1 *hipak1 (const FInteger iv[],
  		   int n) {
    ::hipak1 (getID(), iv, n);
    return this;
  }

  template<class Iterator>
  HBHist1 *hipak1 (FInteger& nid,
  		   Iterator begin,
  		   Iterator end) {
    typename iterator_traits<Iterator>::difference_type n = 0; 
    distance (begin, end, n);  // Don't forget to set n to 0!!!
    n -= 1;
    if (n >= 1) {
      FInteger *bins = new FInteger [n];
      FInteger *out = bins;
      for (Iterator in = begin; in != end; ++in, ++out) {
    	*out = *in;
      }
      hipak1 (nid, bins, n);
      delete[] bins;
    }
    return this;
  }
  
  template<class Iterator>
  HBHist1 *hipak1 (Iterator begin,
  		   Iterator end) {
    FInteger nid = 0;		   
    return hipak1 (nid, begin, end);
  }

  HBHist1 *hipak1 (FInteger& nid,
  		   const vector<FInteger>& v,
  		   int n) {
    return hipak1 (nid, v.begin(), v.begin()+n+1);
  }

  HBHist1 *hipak1 (const vector<FInteger>& v,
  		   int n) {
    return hipak1 (v.begin(), v.begin()+n+1);
  }

  HBHist1 *hipak1 (FInteger& nid,
  		   const vector<FInteger>& v) {
    return hipak1 (nid, v.begin(), v.end());
  }

  HBHist1 *hipak1 (const vector<FInteger>& v) {
    return hipak1 (v.begin(), v.end());
  }
    
  HBHist1 *hpak (const FReal conten[]) {
    ::hpak (getID(), conten);
    return this;
  }
  
  template<class Iterator>
  HBHist1 *hpak (Iterator begin,
  		 Iterator end) {
    int n = noColumns();
//    cerr << "hpak<Iterator>: n = " << n << endl;
    if (n >= 1) {
      FReal *bins = new FReal [n];
      Iterator in = begin;
      FReal *out = bins;
      FReal *outend = bins + n;
      for (; (in != end) && (out != outend); ++in, ++out) {
    	*out = *in;
      }
      for (; out != outend; ++out) {
    	*out = 0;
      }
      hpak (bins);
      delete[] bins;
    }
    return this;
  }
  
  template<class T>
  HBHist1 *hpak (vector<T> v) {
    return hpak (v.begin(), v.end());
  }
  
  HBHist1 *hpake (const FReal conten[]) {
    ::hpake (getID(),conten);
    return this;
  }
  
  template<class Iterator>
  HBHist1 *hpake (Iterator begin,
  		 Iterator end) {
//    cerr << "hpake<Iterator> called.\n";
    int n = noColumns();
//    cerr << "n = " << n << endl;
    if (n >= 1) {
      FReal *bins = new FReal [n];
      Iterator in = begin;
      FReal *out = bins;
      FReal *outend = bins + n;
      for (; (in != end) && (out != outend); ++in, ++out) {
    	*out = *in;
//    	cerr << "hpake<Iterator>: bins["<< out-bins << "] = " << *out << endl;
      }
      for (; out != outend; ++out) {
    	*out = 0;
//    	cerr << "hpake<Iterator>: bins["<< out-bins << "] = " << *out << endl;
      }
      hpake (bins);
      delete[] bins;
    }
    return this;
  }
  
  template<class T>
  HBHist1 *hpake (vector<T> v) {
//    cerr << "hpake<vector<T> > called.\n";
    return hpake (v.begin(), v.end());
  }
  
  static HBHist1 *hrin (int theID, int icycle = 999, int ioffset = 0) {
    return dynamic_cast<HBHist1 *> (HBHisto::hrin (theID, icycle, ioffset));
  }

  static HBHist1 *getHBHist1 (int theID);
  
  virtual void hfill (float x, float weight = 1);
  
  FReal hi (FInteger i) {
    return ::hi (getID(), i);
  }
  
  FReal hie (FInteger i) {
    return ::hie (getID(), i);
  }
  
  FReal hix (FInteger i) {
    return ::hix (getID(), i);
  }
  
  FInteger hxi (FReal x) {
    return ::hxi (getID(), x);
  }
  
  virtual void hbar();
  
  HBHist1Contents contents() {
    return HBHist1Contents (*this);
  }
  
  HBHist1Errors errors() {
    return HBHist1Errors (*this);
  }
  
  int noColumns() {
//    return ::hxi (getID(), numeric_limits<FReal>::max ()) - 1;
//    return ::hxi (getID(), FLT_MAX) - 1; // above would be nicer, needs Rogue Wave's <limits>
    return ::hxi (getID(), 9e37) - 1; // above would be nicer, needs Rogue Wave's <limits>
  }
  
protected:

private:
  const HBHist1& operator= (const HBHist1&);      // no assignment
  
  HBHist1 (int theID, 
           const FCharacter *chtitle, 
           int nx, 
           float xmi,
           float xma,
           float vmx = 0);

  HBHist1 (int theID, 
           const FCharacter *chtitle, 
           int nx, 
           float xmi,
           float xma,
           FReal fun (const FReal&));

  HBHist1 (int theID, 
           const FCharacter *chtitle, 
           int nx, 
           const float xbins[],
           float vmx = 0);

  HBHist1 (const HBHist1 &oldHist1,
           int theID, 
           const FCharacter *chtitle);
           
  HBHist1 (int theID) 
    // generates an Object for a one-dimensional histogram already in 
    // memory
  : HBHisto (theID) {
    cerr << "HBHist1::HBHist1: New HBHist1 object has been created, id=" << getID() << "\n";
    cerr << "Histogram was already in memory.\n";
  }

};
class HBHist2;                          
//**********************************************************************
template<class F> 
class HBFunctionAdaptor2: private HBFunctionAdaptorBase {
//**********************************************************************
public:
  static void setFunction (const F *f) {HBFunctionAdaptorBase::setFunction (f);}
  static void setFunction (const F& f) {HBFunctionAdaptorBase::setFunction (&f);}
  static FReal fun (const FReal& x, const FReal& y) {
    return static_cast<FReal> ((*HBFunctionAdaptorBase::getFunction((F *)(0)))(x, y));
  }
};

//**********************************************************************
class HBHist2: public HBHisto {
//**********************************************************************
public:
  virtual HBHist2 *hcopy (int newID, 
                          const FCharacter *chtitle);

  static HBHist2 *hbook2 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          int ny, 
                          float ymi,
                          float yma,
                          float vmx = 0);
                          
  static HBHist2 *hbfun2 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          int ny, 
                          float ymi,
                          float yma,
                          FReal fun (const FReal&, const FReal&));
                         
  template<class F>                        
  static HBHist2 *hbfun2 (int theID, 
                          const FCharacter *chtitle, 
                          int nx, 
                          float xmi,
                          float xma,
                          int ny, 
                          float ymi,
                          float yma,
                          F fobject) {
    if (getHBEntity (theID) != 0) {
      cerr << "HBHist2::hbfun2: HBHist2 cannot be created, id=" << theID << "\n";
      return 0; 	// ID already in use
    } else {
      cerr << "HBHist2::hbfun2: HBHist2 will be created, id=" << theID << "\n";
      HBFunctionAdaptor2<F>::setFunction (fobject);     // this function sets static
                                                        // pointer F in class Function<F>
                                                        // to fobject, which is the then
                                                        // called by Function<F>::fun
      return new HBHist2 (theID, chtitle, nx, xmi, xma, ny, ymi, yma, &HBFunctionAdaptor2<F>::fun);
    }
  }
                          
  static HBHist2 *getHBHist2 (int theID);
                          
  static HBHist2 *makeObject (int theID);
  
  static HBHist2 *hrin (int theID, int icycle = 999, int ioffset = 0) {
    return dynamic_cast<HBHist2 *> (HBHisto::hrin (theID, icycle, ioffset));
  }
  
  virtual void hfill (float x, float y, float weight = 1);
  
  virtual void hbar();
  virtual void hbary();
  virtual void hbar2();
  
  inline void hscale (float factor = 0);
  inline void hrndm2 (float &x, float &y);
protected:

private:
  const HBHist2& operator= (const HBHist2&);      // no assignment
  
  HBHist2 (int theID, 
           const FCharacter *chtitle, 
           int nx, 
           float xmi,
           float xma,
           int ny, 
           float ymi,
           float yma,
           float vmx = 0);

  HBHist2 (int theID, 
           const FCharacter *chtitle, 
           int nx, 
           float xmi,
           float xma,
           int ny, 
           float ymi,
           float yma,
           FReal fun (const FReal&, const FReal&));

  HBHist2 (const HBHist2 &oldHist2,
           int theID, 
           const FCharacter *chtitle);
           
  HBHist2 (int theID) 
    // generates an Object for a two-dimensional histogram already in 
    // memory
  : HBHisto (theID) {
    cerr << "HBHist2::HBHist2: New HBHist2 object has been created, id=" << getID() << "\n";
    cerr << "Histogram was already in memory.\n";
  }

};

//**********************************************************************
// Definitions of inline functions follow
//**********************************************************************

//**********************************************************************
// class HBEntity
//**********************************************************************

inline int HBEntity::getID() const {
  return id;
}

//**********************************************************************
// class HBEntity
//**********************************************************************

void HBEntity::htitle (const FCharacter *chgtit) {
  ::htitle (chgtit);
}

//**********************************************************************
// class HBHisto
//**********************************************************************

void HBHisto::hprint() const {
  ::hprint (getID());
}

void HBHisto::histdo() {
  ::histdo ();
}

//**********************************************************************
// template<class SomeHBHist1Iterator>
// class HBHist1IteratorBase <SomeHBHist1Iterator>
//**********************************************************************

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator&	   HBHist1IteratorBase<SomeHBHist1Iterator>::operator+= (int offset)
{
  theColumn += offset;
  return reinterpret_cast<SomeHBHist1Iterator&> (*this);
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator&	   HBHist1IteratorBase<SomeHBHist1Iterator>::operator-= (int offset)
{
  theColumn -= offset;
  return reinterpret_cast<SomeHBHist1Iterator&> (*this);
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator HBHist1IteratorBase<SomeHBHist1Iterator>::operator+ (int offset)
{
  return SomeHBHist1Iterator (*this) += offset;
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator HBHist1IteratorBase<SomeHBHist1Iterator>::operator- (int offset)
{
  return SomeHBHist1Iterator (*this) -= offset;
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator&	   HBHist1IteratorBase<SomeHBHist1Iterator>::operator++()    //prefix increment
{
  ++theColumn;
  return reinterpret_cast<SomeHBHist1Iterator&> (*this);
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator&	   HBHist1IteratorBase<SomeHBHist1Iterator>::operator--()    //prefix decrement
{
  --theColumn;
  return reinterpret_cast<SomeHBHist1Iterator&> (*this);
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator HBHist1IteratorBase<SomeHBHist1Iterator>::operator++ (int) //postfix increment
{
  const SomeHBHist1Iterator oldValue (*this);
  ++theColumn;
  return oldValue;
}

template<class SomeHBHist1Iterator>
SomeHBHist1Iterator HBHist1IteratorBase<SomeHBHist1Iterator>::operator-- (int) //postfix decrement
{
  const SomeHBHist1Iterator oldValue (*this);
  --theColumn;
  return oldValue;
}
  
template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator== (int index)
{ 
  return theColumn == index;
}

template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator!= (int index)
{ 
  return theColumn != index;
}

template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator< (int index)
{ 
  return theColumn < index;
}

template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator<= (int index)
{ 
  return theColumn <= index;
}

template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator> (int index)
{ 
  return theColumn > index;
}

template<class SomeHBHist1Iterator>
bool         HBHist1IteratorBase<SomeHBHist1Iterator>::operator>= (int index)
{ 
  return theColumn >= index;
}

template<class SomeHBHist1Iterator>
inline int  operator-  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn - rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator== (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn == rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator!= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn != rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator<  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn < rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator<= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn <= rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator>  (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs,  
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn > rhs.theColumn;
}

template<class SomeHBHist1Iterator>
inline bool operator>= (const HBHist1IteratorBase<SomeHBHist1Iterator>& lhs, 
                        const HBHist1IteratorBase<SomeHBHist1Iterator>& rhs)
{
  return lhs.theColumn >= rhs.theColumn;
}
  
//**********************************************************************
// class HBHist1ContentsIterator 
//**********************************************************************

HBHist1ContentsIterator::value_type HBHist1ContentsIterator::operator*() {
  return theHistogram.hi(theColumn);
}

HBHist1ContentsIterator::value_type HBHist1ContentsIterator::operator[](difference_type offset) {
  return *(HBHist1ContentsIterator (*this) += offset);
}
  
//**********************************************************************
// class HBHist1Contents 
//**********************************************************************

HBHist1Contents::value_type HBHist1Contents::operator[](difference_type offset) {
  return begin()[offset];
}

HBHist1ContentsIterator HBHist1Contents::begin() {
  return HBHist1ContentsIterator (theHistogram, 1);
}

HBHist1ContentsIterator HBHist1Contents::end() {
  return HBHist1ContentsIterator (theHistogram, theHistogram.noColumns()+1);
}

HBHist1Contents::difference_type HBHist1Contents::size() {
  return theHistogram.noColumns();
}

template<class Container>
HBHist1Contents& HBHist1Contents::operator= (Container rhs) {
  theHistogram.hpak (rhs.begin(), rhs.end());
  return *this;
}

//**********************************************************************
// class HBHist1ErrorsIterator 
//**********************************************************************

HBHist1ErrorsIterator::value_type HBHist1ErrorsIterator::operator*() {
  return theHistogram.hie(theColumn);
}

HBHist1ErrorsIterator::value_type HBHist1ErrorsIterator::operator[](difference_type offset) {
  return *(HBHist1ErrorsIterator (*this) += offset);
}
  
//**********************************************************************
// class HBHist1Errors 
//**********************************************************************

HBHist1Errors::value_type HBHist1Errors::operator[](difference_type offset) {
  return begin()[offset];
}

HBHist1ErrorsIterator HBHist1Errors::begin() {
  return HBHist1ErrorsIterator (theHistogram, 1);
}

HBHist1ErrorsIterator HBHist1Errors::end() {
  return HBHist1ErrorsIterator (theHistogram, theHistogram.noColumns()+1);
}

HBHist1Errors::difference_type HBHist1Errors::size() {
  return theHistogram.noColumns();
}

template<class Container>
HBHist1Errors& HBHist1Errors::operator= (Container rhs) {
  theHistogram.hpake (rhs.begin(), rhs.end());
  return *this;
}


//**********************************************************************
// class HBHist1
//**********************************************************************

float HBHist1::hrndm1() const {
  return ::hrndm1 (getID());
}                        

//**********************************************************************
// class HBHist2
//**********************************************************************

void HBHist2::hscale(float factor) {
  ::hscale (getID(), factor);
}                        

void HBHist2::hrndm2(float &x, float&y) {
  ::hrndm2 (getID(), x, y);
}                        

class HBFile;
//**********************************************************************
class HBFile {
//**********************************************************************
public:
  
  typedef map<int, HBFile *> MapType;
  typedef MapType::iterator MapIterator;

  static HBFile *hropen (int lun, string chtop, string chfile, string chopt, int& lrec) {
    HBFile *file = 0;
    if (fileMap [lun] != 0) {
      cerr << "HBFile::hropen: cannot open file " << chfile.c_str() 
           << " on unit " << lun << ", unit occupied.";
    }
    else {
      FInteger istat = 0;
      file = new HBFile (lun, chtop, chfile, chopt, lrec, istat);
      if (istat != 0) {
        cerr << "HBFile::hropen: cannot open file " << chfile.c_str() 
             << " on unit " << lun << ", error " << istat << " from hropen.";
        delete file;
      }
    }
    return file;
  }
  
  static HBFile *hropen (int lun, string chtop, string chfile, string chopt = " ") {
    int lrec = 1024;
    return hropen (lun, chtop, chfile, chopt, lrec);
  }
  
  ~HBFile() {
    ::hrend (topdir);
    fortclose (unit);
    fileMap [unit] = 0;
  }
  
protected:  
  HBFile (int lun, string chtop, string chfile, string chopt, int& lrec, FInteger& istat)
  : unit (lun), topdir (chtop), filename (chfile), options (chopt), reclen (lrec) {
    ::hropen (unit, topdir, filename, options, reclen, istat);
    lrec = reclen;
    fileMap [unit] = this;
  }

private:
  int unit;
  string topdir;
  string filename;
  string options;
  FInteger reclen;
  
  static MapType fileMap;  
};

#endif /* #ifndef HBOOKOO_H */
