#include "ftypes.h"
//#include "hbook.h"  Stefan has already another hbook.h ....
#include "hbookwrap.h"
#include <cmath>

template<class Op>
void hopera (FInteger id1, 
             Op op, 
             FInteger id2, 
             FInteger id3,
             FReal c1 = 1,
             FReal c2 = 1) {
// This hopera expects a function object that overloads the 
// () operator like this:
//   FReal operator() (FReal x, FReal y)
// to yield the new result, and
// ALSO defines an operator () like this:
//   FReal (FReal x, FReal xerr, FReal y, FReal yerr)
// to yield the desired error

  FCharacter chtitl [81];
  FInteger nx;
  FReal xmi;
  FReal xma;
  FInteger ny;
  FReal ymi;
  FReal yma;
  FInteger nwt;
  FInteger loc;
  
  hgive (id1, chtitl, nx, xmi, xma, ny, ymi, yma, nwt, loc); 

// works only for 1-dim histos at the moment!
  int n = nx;
  
  if (!hexist (id3)) {
    hcopy (id1, id3);
  }
  
  FReal *conten = new FReal [n];
  FReal *errors = new FReal [n];
  
  for (int i = 1; i <= n; ++i) {
    FReal x1    = c1*hi (id1, i);
    FReal x2    = c2*hi (id2, i);
    FReal xerr1 = c1*hie (id1, i);
    FReal xerr2 = c2*hie (id2, i);
    
    conten [i-1] = op (x1, x2);
    errors [i-1] = op (x1, xerr1, x2, xerr2);
  }
  hpak  (id3, conten);
  hpake (id3, errors);
  
  delete[] conten;
  delete[] errors;
}
  
template<class Op>
void hopera (FInteger id1, 
             Op op, 
             FInteger id2, 
             FInteger id3,
             FReal c1,
             FReal c2,
             FReal h) {
// This hopera expects a function object that overloads the 
// () operator like this:
//   FReal operator() (FReal x, FReal y);
// if h>0, errors are calculated by varying the contents of the 
// input histograms by h*error, otherwise the resulting error is 0

  FCharacter chtitl [81];
  FInteger nx;
  FReal xmi;
  FReal xma;
  FInteger ny;
  FReal ymi;
  FReal yma;
  FInteger nwt;
  FInteger loc;
  
  hgive (id1, chtitl, nx, xmi, xma, ny, ymi, yma, nwt, loc); 

// works only for 1-dim histos at the moment!
  int n = nx;
  
  if (!hexist (id3)) {
    hcopy (id1, id3);
  }
  
  FReal *conten = new FReal [n];
  FReal *errors = new FReal [n];
  
  for (int i = 1; i <= n; ++i) {
    FReal x1    = c1*hi (id1, i);
    FReal x2    = c2*hi (id2, i);
    FReal xerr1 = h*c1*hie (id1, i);
    FReal xerr2 = h*c2*hie (id2, i);
    
    conten [i-1] = op (x1, x2);
    if (h > 0) {
      FReal e1 = op (x1+xerr1, x2) - op (x1-xerr1, x2);
      FReal e2 = op (x1, x2+xerr2) - op (x1, x2+xerr2);
      errors [i-1] = sqrt (e1*e1 + e2*e2)*(0.5/h);
    }
    else {
      errors [i-1] = 0;
    }
  }
  hpak  (id3, conten);
  hpake (id3, errors);
  
  delete[] conten;
  delete[] errors;
}
  
  
  
